#!/bin/bash

#----------------------------------------------------------------------------
# Copyright (c) 2010, Intel Corporation.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 
# 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#
#
#----------------------------------------------------------------------------

#
# EMGD Driver Install Script
#

###############################################################################
#
# Globals
#
###############################################################################

PRODUCT="EMGD"
PVR_VERSION=".1.5.15.3226"

# Argument vars
VERBOSE=0
UNINSTALL=0
REINSTALL=0
SILENT=0
MAXARGS=3

# Directory where this ./install.sh lives
INSTALL_DIR=$(pwd)

# ISSUE File
ISSUEFILE="/etc/issue"

# Distribution Names
DISTRO_FEDORA14="Fedora 14" 
DISTRO_MEEGO1_2="MeeGo 1.2"
DISTRO_UNKNOWN="Unknown"
DISTRO=$DISTRO_UNKNOWN

# Distribution folder names
DISTRO_F14_FOLDER="F14"
DISTRO_FOLDER=$DISTRO_UNKNOWN

# X-Server Names
XVERSION_1_9="Xorg-xserver-1.9.0"
XVERSION=$XSERVER_UNKNOWN

# Command Name Paths
LSPCI_CMD="/sbin/lspci"
TAR_CMD="/bin/tar"
MAKE_CMD="/usr/bin/make"
LSMOD_CMD="/sbin/lsmod"
RMMOD_CMD="/sbin/rmmod"
INSMOD_CMD="/sbin/insmod"
MODPROBE_CMD="/sbin/modprobe"

# Error codes
E_OK=0
E_TOOMANYARGS=1
E_NODIR=2
E_NOROOT=3
E_NOCMD=4
E_NOTSUPPORTED=5
E_SEENABLED=6
E_NOKERHEADERS=7
E_BUILDERR=8

# Intel vendor id
INTEL_VEN="8086"
# Device id for US15 and TNC
EMGD_US15_DEV=( '8108' )
EMGD_TNC_DEV=( '4108' '8182' '4114' )

# SELinux
DISABLE_SELINUX="y";

# root directory path to source files, needed to build DRM module
EGD_CFG="drv"

# Source file
SOURCE_FILE="emgd_drm.tgz"

###############################################################################
#
# Functions
#
###############################################################################

#------------------------------------------------------------------------------
# Usage text
# 	Prints usage text info for this script.
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
show_usage() {
	# Print usage text
	echo "Usage: ./install.sh [OPTION]..."
	echo "$PRODUCT installation script."
	echo "  -h    this usage text"
	echo "  -r    reinstall"
	echo "  -u    uninstall"
	echo "  -v    verbose output"

	return $E_OK
}

#------------------------------------------------------------------------------
# Verbose
#	Ouputs string if VERBOSE flag is set.
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
verbose() {
	# Check if VERBOSE is set and if we have an argument to the function.
	if [[ $VERBOSE -eq 1 && $# -eq 1 ]]; then
		# Output the text
		echo "  $1"
	fi

	return $E_OK
}

#------------------------------------------------------------------------------
# install_file
#	Installs file if file exists
#	Input:  file
#		destination file name
#	Output: none
#	Returns: none
#------------------------------------------------------------------------------
install_file() {
	
	if [ ! -e $1 ]; then
		verbose "The file $1 not found, skipping." 
	else 
		verbose "Installing the file $1 to $2."
		# delete the original first
		rm -f $2
		# now copy the file
		cp -f $1 $2
	fi

	return;
}

#------------------------------------------------------------------------------
# install_link
#	creates a link if file exists
#	Input:  file
#		destination file name
#	Output: none
#	Returns: none
#------------------------------------------------------------------------------
install_link()
{

	if [ ! -e $1 ]; then
		verbose "The file $1 not found, skipping." 
	else 
		verbose "Creating the link of file $1 to $2."
		# delete the original first
		rm -f $2
		# now copy the file
		ln -sfv $1 $2
	fi

	return;
}

#------------------------------------------------------------------------------
# Save Remove
#	Removes files only if it exists
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
safe_rm() {
	if [[ $# -eq 1 ]]; then
		if [[ -e $1 ]]; then
			rm -v $1
		else
			verbose "$FUNCNAME file $1 does not exist."
		fi
	fi

	return $E_OK
}

#------------------------------------------------------------------------------
# Pre Install
#	Check to make sure all destination directories exist before starting copy.
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK or E_NODIR
#------------------------------------------------------------------------------
pre_install() {
	verbose "$FUNCNAME Default Installation entry"

	echo ""
	echo "Installing $PRODUCT on "$DISTRO
	echo ""

	# Check to make sure lspci command exists
	if [ ! -e $LSPCI_CMD ]; then
		echo "The command $LSPCI_CMD not found, exiting."
		exit $E_NOCMD
	fi
	
	# Check to make sure tar command exists
	if [ ! -e $TAR_CMD ]; then
		echo "The command $TAR_CMD not found, exiting."
		exit $E_NOCMD
	fi

	# Check to make sure make command exists
	if [ ! -e $MAKE_CMD ]; then
		echo "The command $MAKE_CMD not found, exiting."
		exit $E_NOCMD
	fi

	# Check to make sure MODPROBE command exists
	if [ ! -e $MODPROBE_CMD ]; then
		echo "The command $MODPROBE_CMD not found, exiting."
		exit $E_NOCMD
	fi

	# Check for driver install directory
	loc_driver_install_dir="/usr/lib/xorg/modules/drivers/"

	if [ ! -d $loc_driver_install_dir ]; then
		echo "The directory $loc_driver_install_dir must exist to install $PRODUCT."
		return $E_NODIR
	fi

	# Check for dri install directory
	loc_dri_dir="/usr/lib/dri/"

	if [ ! -d $loc_dri_dir ]; then
		echo "The directory $loc_dri_dir must exist to install $PRODUCT."
		return $E_NODIR
	fi
	verbose "$FUNCNAME The directory $loc_dri_dir exists."

	# Check for man pages install directory
	man_install_dir="/usr/share/man/man4/"

	if [ ! -d $man_install_dir ]; then
		echo "The directory $man_install_dir must exist to install $PRODUCT."
		return $E_NODIR
	fi
	verbose "$FUNCNAME The directory $man_install_dir exists."

	# Check for man emgd_escape.3x.gz pages install directory
	man3x_install_dir="/usr/share/man/man3x/"

	if [ ! -d $man3x_install_dir ]; then
		echo "The directory $man3x_install_dir must exist to install $PRODUCT."
		return $E_NODIR
	fi
	verbose "$FUNCNAME The directory $man3x_install_dir exists."

	# Check for usr/lib directory
	lib_install_dir="/usr/lib/"

	if [ ! -d $lib_install_dir ]; then
		echo "The directory $lib_install_dir must exist to install $PRODUCT."
		return $E_NODIR
	fi
	verbose "$FUNCNAME The directory $lib_install_dir exists."

	# Check for the /etc directory
	cfg_install_dir="/etc/"

	if [ ! -d $cfg_install_dir ]; then
		echo "The directory $cfg_install_dir must exist to install $PRODUCT."
		return $E_NODIR
	fi
	verbose "$FUNCNAME The directory $cfg_install_dir exists."

    # Ensure that the khronos header directories exists
    inc_install_dir="/usr/include"
    mkdir -p $inc_install_dir/EGL
    mkdir -p $inc_install_dir/GLES
    mkdir -p $inc_install_dir/GLES2
    mkdir -p $inc_install_dir/KHR
    mkdir -p $inc_install_dir/VG

	# Check that the KERNELDIR variable is set
	# If KERNELDIR is not set, then check for the local
	if [[ $KERNELDIR == "" ]]; then
		KERNEL=`uname -r`
		if [ ! -d "/lib/modules/$KERNEL/build" ]; then
			echo "There were no Kernel Headers found!  "
			echo -ne "To build the EMGD DRM module the "
			echo -ne "KERNELDIR environment variable will need "
			echo -ne "to be set for the desired kernel or install "
			echo -ne "the $KERNEL Headers."
			return $E_NOKERHEADERS
		fi
		KERNELDIR="/lib/modules/$KERNEL/build"
	fi
	verbose "$FUNCNAME The variable KERNELDIR is set to $KERNELDIR."
	
	return $E_OK
}

#------------------------------------------------------------------------------
# Copy files (Video)
#	Copy video driver files
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
copy_files_video() { 

	# Copy Video Decode files
	echo "Copying Video Acceleration libraries"
	cd "$INSTALL_DIR/../$DISTRO_FOLDER/driver/$XVERSION"
	install_file emgd_drv_video.so ${loc_driver_install_dir}emgd_drv_video.so

	# Copy the LIBVA files only for Fedora 14
	if [[ $DISTRO = $DISTRO_FEDORA14 ]]; then
		install_file libva.so.1.0.1 ${lib_install_dir}libva.so.1.0.1
		install_file libva-x11.so.1.0.1 ${lib_install_dir}libva-x11.so.1.0.1
	fi

	return $E_OK
}


#------------------------------------------------------------------------------
# Create links (Video)
#	Create video driver links
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
create_links_video() {
	
	# Create Video Decode Links
	echo "Creating Video Acceleration library links"
	cd "$lib_install_dir"

	if [[ $DISTRO = $DISTRO_FEDORA14 ]]; then
		install_link libva.so.1.0.1 libva.so
		install_link libva.so.1.0.1 libva.so.1
		install_link libva-x11.so.1.0.1 libva-x11.so
		install_link libva-x11.so.1.0.1 libva-x11.so.1
	fi

	return $E_OK
}


#------------------------------------------------------------------------------
# Copy files
#	Copy driver files
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
copy_files() { 

	# Start copy output on a new line
	echo ""

	# Copy the Driver
	echo "Copying the driver"
	cd "$INSTALL_DIR/../$DISTRO_FOLDER/driver/$XVERSION"
	install_file emgd_drv.so ${loc_driver_install_dir}emgd_drv.so
	install_file libXegd_escape.so.2.0.0 ${lib_install_dir}libXegd_escape.so.2.0.0
	install_file emgd_dri.so ${loc_dri_dir}emgd_dri.so

	install_file libGLES_CM.so ${lib_install_dir}libGLES_CM.so$PVR_VERSION
	install_file libGLESv2.so ${lib_install_dir}libGLESv2.so$PVR_VERSION
	install_file libemgdglslcompiler.so ${lib_install_dir}libemgdglslcompiler.so$PVR_VERSION
	install_file libOpenVG.so ${lib_install_dir}libOpenVG.so$PVR_VERSION
	install_file libOpenVGU.so ${lib_install_dir}libOpenVGU.so$PVR_VERSION
	install_file libEMGDegl.so ${lib_install_dir}libEMGDegl.so$PVR_VERSION
	install_file libEGL.so ${lib_install_dir}libEGL.so$PVR_VERSION
	install_file libEMGD2d.so ${lib_install_dir}libEMGD2d.so$PVR_VERSION
	install_file libemgdPVR2D_DRIWSEGL.so ${lib_install_dir}libemgdPVR2D_DRIWSEGL.so$PVR_VERSION
	install_file libemgdsrv_um.so ${lib_install_dir}libemgdsrv_um.so$PVR_VERSION
	install_file libemgdsrv_init.so ${lib_install_dir}libemgdsrv_init.so$PVR_VERSION
	install_file libEMGDScopeServices.so ${lib_install_dir}libEMGDScopeServices.so$PVR_VERSION
	install_file libEMGDOGL.so ${lib_install_dir}libEMGDOGL.so$PVR_VERSION
	install_file libSGXSGL.so ${lib_install_dir}libSGXSGL.so$PVR_VERSION

	# Copy the Man Pages
	cd "$INSTALL_DIR/../$DISTRO_FOLDER/driver/$XVERSION"
	
	install_file emgd.4.gz ${man_install_dir}emgd.4.gz
		
	install_file emgd_escape.3x.gz ${man3x_install_dir}emgd_escape.3x.gz

	# Copy Video Decode files
	copy_files_video

	# Copy emgd_info (only for internal tests)
	lib_internal_dir="$INSTALL_DIR/Internal/"	
	if [[ $SILENT -eq 1 ]]; then
		if [ ! -d $lib_install_dir ]; then
			echo "The directory $lib_internal_dir does not exist."
			echo "emgdinfo will not be installed."
		else
			verbose "$FUNCNAME The directory $lib_internal_dir exists."
			cd "$lib_internal_dir"
			install_file libemgdinfo.so ${lib_install_dir}libemgdinfo.so
		fi
	fi

	# Copy the 3D config file
	install_file powervr.ini ${cfg_install_dir}powervr.ini

    # Copy the khronos development header files
    cd "$INSTALL_DIR/../common/headers/EGL"
    install_file egl.h ${inc_install_dir}/EGL/egl.h
    install_file eglext.h ${inc_install_dir}/EGL/eglext.h
    install_file eglplatform.h ${inc_install_dir}/EGL/eglplatform.h
    cd "$INSTALL_DIR/../common/headers/GLES"
    install_file gl.h ${inc_install_dir}/GLES/gl.h
    install_file glext.h ${inc_install_dir}/GLES/glext.h
    install_file glplatform.h ${inc_install_dir}/GLES/glplatform.h
    cd "$INSTALL_DIR/../common/headers/GLES2"
    install_file gl2.h ${inc_install_dir}/GLES2/gl2.h
    install_file glext2.h ${inc_install_dir}/GLES2/glext2.h
    install_file glplatform2.h ${inc_install_dir}/GLES2/glplatform2.h
    cd "$INSTALL_DIR/../common/headers/KHR"
    install_file khrplatform.h ${inc_install_dir}/KHR/khrplatform.h
    cd "$INSTALL_DIR/../common/headers/VG"
    install_file openvg.h ${inc_install_dir}/VG/openvg.h
    install_file vgext.h ${inc_install_dir}/VG/vgext.h
    install_file vgplatform.h ${inc_install_dir}/VG/vgplatform.h
    install_file vgu.h ${inc_install_dir}/VG/vgu.h

	return $E_OK
}

#------------------------------------------------------------------------------
# Create links
#	Copy driver files
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
create_links() {

	# For re-install/upgrade unlink the libraries
	cd "$lib_install_dir"
	if [[ $REINSTALL -eq 1 ]]; then
		unlink libXegd_escape.so
		unlink libXegd_escape.so.2
		unlink libva.so
		unlink libva.so.1
		unlink libva-x11.so
		unlink libva-x11.so.1
	fi

	# Create Escape Libraries Links
	echo "Creating Escape library links"
	install_link libXegd_escape.so.2.0.0 libXegd_escape.so
	install_link libXegd_escape.so.2.0.0 libXegd_escape.so.2

	# Create PowerVR links
	install_link libGLES_CM.so$PVR_VERSION libGLES_CM.so
	install_link libGLESv2.so$PVR_VERSION libGLESv2.so
	install_link libemgdglslcompiler.so$PVR_VERSION libemgdglslcompiler.so
	install_link libOpenVG.so$PVR_VERSION libOpenVG.so
	install_link libOpenVGU.so$PVR_VERSION libOpenVGU.so
	install_link libEMGDegl.so$PVR_VERSION libEMGDegl.so
	install_link libEGL.so$PVR_VERSION libEGL.so
	install_link libEMGD2d.so$PVR_VERSION libEMGD2d.so
	install_link libemgdPVR2D_DRIWSEGL.so$PVR_VERSION libemgdPVR2D_DRIWSEGL.so
	install_link libemgdsrv_um.so$PVR_VERSION libemgdsrv_um.so
	install_link libemgdsrv_init.so$PVR_VERSION libemgdsrv_init.so
	install_link libEMGDScopeServices.so$PVR_VERSION libPVRScopeServices.so
	install_link libEMGDOGL.so$PVR_VERSION libEMGDOGL.so
	install_link libSGXSGL.so$PVR_VERSION libSGXSGL.so
	
	# Create Video Decode Links
	create_links_video

	# Create Open GL links
	echo "Creating OpenGL library links"
	install_link libEGL.so$PVR_VERSION libEGL.so.1
	install_link libGLES_CM$PVR_VERSION libGLES_CM.so.1
	install_link libEGL.so$PVR_VERSION libGLESv1.so
	install_link libGLESv2$PVR_VERSION libGLESv2.so.2

	return $E_OK
}


#------------------------------------------------------------------------------
# EMGD GUI
#	Adds EMGD GUI shortcut to the Applications->Systems menu
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
emgd_gui() {

	# Install EMGDGUI shortcut
	cd "$INSTALL_DIR/../$DISTRO_FOLDER/Utilities"
	local emgdgui_install_dir="/usr/share/applications"
	local emgdgui_filename="$emgdgui_install_dir/emgdgui.desktop"
	local emgdgui_icon="/usr/share/icons/gnome/scalable/devices/video-display.svg"

	if [[ -e $emgdgui_install_dir ]]; then

		echo "Creating EMGDGUI GNOME shortcut"
		if [[ ! -x ./emgdgui ]]; then
			chmod +x emgdgui
		fi

		# Copy the EMGDGUI application
		install_file emgdgui "/usr/bin/emgdgui"

		# Create GNOME shortcut
		echo "[Desktop Entry]" > $emgdgui_filename 
		echo "Name=EMGD GUI Utility" >> $emgdgui_filename 
		echo "Comment=EMGD Configuration" >> $emgdgui_filename 
		echo "Exec=/usr/bin/emgdgui" >> $emgdgui_filename 
		echo "Icon=$emgdgui_icon" >> $emgdgui_filename  
		echo "Type=Application" >> $emgdgui_filename 
		echo "Encoding=UTF-8" >> $emgdgui_filename  
		echo "Categories=Application;System;" >> $emgdgui_filename  

	fi

	verbose "$FUNCNAME Default Installation exit"

	return $E_OK
}

#------------------------------------------------------------------------------
# Install EMGD DRM
#	1. EMGD DRM never installed (exists=0 & loaded=0)
#    		build, copy, insmod the DRM
#  	2. EMGD DRM exists 
#		ask user to build & copy, then insmod 
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#                E_NODIR
#                E_BUILDERR
#------------------------------------------------------------------------------
drm() {
	local rebuild="y"
	local reload=""
	local extractsources="y"

	echo "Building and Installing the EMGD DRM Module"
	cd "$INSTALL_DIR/../common/drm"

	# If Reinstall, then check if directories still exists
	if [[ $SILENT -eq 0 ]]; then

		# Check to see if the EMGD DRM module already exists
		EMGD_EXISTS=`ls | grep "emgd.ko"`

		verbose "$FUNCNAME Checking if EMGD DRM already exists"
		if [[ $EMGD_EXISTS == "emgd.ko" ]]; then
			echo ""
			echo -ne "EMGD DRM module already exists.  Do you want to"
			echo -ne " rebuild and reinstall another EMGD DRM module? [y/N]"
			read rebuild
		fi
	fi

	if [[ $rebuild == "Y" || $rebuild == "y" ]]; then
		# Source Directories exists, skip to install
		EGD_ROOT="$INSTALL_DIR/../common/drm/emgd"
		
		# Check if user wants to extract sources or use existing
		# sources directory
		if [[ $SILENT -eq 0 ]]; then
			if [ -d $EGD_ROOT ]; then
				echo ""
				echo -ne "The directory $EGD_ROOT currently "
				echo "exists."
				echo -ne "Do you want to extract the "
				echo -ne "sources from emgd_drm.tgz? [y/N]?"
				read extractsources
			fi
		fi
		
		if [[ $extractsources == "Y" || $extractsources == "y" ]]; then
			# Extract source directories
			echo "Extracting the sources"
			$TAR_CMD -xvzf $SOURCE_FILE 
		fi
		
		if [ ! -d $EGD_ROOT ]; then
			echo -ne "The directory $EGD_ROOT does not exist.  "
			echo -ne "The directory must exist for EMGD to install "
			echo -ne "and build the EMGD DRM module."
			return $E_NODIR
		fi

		# Perform a Make 
 		$MAKE_CMD 
	
		if [ ! -e "emgd.ko" ]; then
			echo -ne "The build process of the EMGD DRM module "
			echo -ne "has failed. "
			return $E_BUILDERR
		fi
	fi

	# Perform a Make Install
	$MAKE_CMD install

	# Check to see if DRM is currently loaded
	DRM_LOADED=`$LSMOD | grep "drm"`
	
	if [[ $DRM_LOADED != "drm" ]]; then
		$MODPROBE_CMD drm
	fi

	# Insert the EMGD DRM module 
	if [[ $DISTRO == $DISTRO_FEDORA14 ]]; then
	
		DRM_KMS_LOADED=`$LSMOD | grep "drm_kms_helper"`
	
		if [[ $DRM_KMS_LOADED != "drm_kms_helper" ]]; then
			$MODPROBE_CMD drm_kms_helper
		fi
	fi

	# Insert the EMGD DRM module 
 	if [[ $MODPROBE_CMD != "" ]]; then
		$MODPROBE_CMD emgd
	else
		$INSMOD_CMD emgd.ko
	fi

	return $E_OK
}

#------------------------------------------------------------------------------
# Default Uninstall
#	Driver installation function designated the 'default' uninstallation.
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK
#------------------------------------------------------------------------------
default_uninstall() {
	verbose "$FUNCNAME Default Uninstallation entry"

	echo "Uninstalling $PRODUCT on "$DISTRO
	echo ""
	
	# Remove Driver
	safe_rm /usr/lib/xorg/modules/drivers/emgd_drv.so
	safe_rm /usr/lib/dri/emgd_dri.so

	# Remove Port Drivers
	safe_rm /usr/lib/libGLES_CM.so$PVR_VERSION
	safe_rm /usr/lib/libGLESv2.so$PVR_VERSION
	safe_rm /usr/lib/libemgdglslcompiler.so$PVR_VERSION
	safe_rm /usr/lib/libOpenVG.so$PVR_VERSION
	safe_rm /usr/lib/libOpenVGU.so$PVR_VERSION
	safe_rm /usr/lib/libEMGDegl.so$PVR_VERSION
	safe_rm /usr/lib/libEGL.so$PVR_VERSION
	safe_rm /usr/lib/libEMGD2d.so$PVR_VERSION
	safe_rm /usr/lib/libemgdPVR2D_BLITWSEGL.so$PVR_VERSION
	safe_rm /usr/lib/libemgdPVR2D_FLIPWSEGL.so$PVR_VERSION
	safe_rm /usr/lib/libemgdPVR2D_LINUXFBWSEGL.so$PVR_VERSION
	safe_rm /usr/lib/libemgdPVR2D_DRIWSEGL.so$PVR_VERSION
	safe_rm /usr/lib/libemgdemgdsrv_um.so$PVR_VERSION
	safe_rm /usr/lib/libemgdsrv_init.so$PVR_VERSION
	safe_rm /usr/lib/libEMGDScopeServices.so$PVR_VERSION
	safe_rm /usr/lib/libEMGDOGL.so$PVR_VERSION
	safe_rm /usr/lib/libSGXSGL.so$PVR_VERSION
	
	# Remove Escape libararies
	safe_rm /usr/lib/libXemgd_escape.so.2.0.0
	safe_rm /usr/lib/libXemgd_escape.so
	safe_rm /usr/lib/libXemgd_escape.so

	# Remove the Man Pages
	safe_rm	/usr/share/man/man4/emgd.4.gz
	safe_rm /var/cache/man/cat4/emgd.4.bz2
	safe_rm	/usr/share/man/man3x/emgd_escape.3x.gz
	

	# Remove EMGDGUI shortcut
	cd "$INSTALL_DIR"
	local emgdgui_install_dir="/usr/share/applications"
	local emgdgui_filename="$emgdgui_install_dir/emgdgui.desktop"

	safe_rm /usr/bin/emgdgui
	safe_rm $emgdgui_filename
	
	# Remove the X version test file
	local xversionFileName=.xversionFile
	if [ -e "$INSTALL_DIR/$xversionFileName" ]; then
		safe_rm $INSTALL_DIR/$xversionFileName
	fi

	# Remove Video Decode
	#safe_rm /usr/lib/libva.so.1.0.1
	#safe_rm /usr/lib/libva-x11.so.1.0.1
	#safe_rm /usr/lib/xorg/modules/drivers/emgd_drv_video.so

    # Remove development headers
    safe_rm /usr/include/EGL/egl.h
    safe_rm /usr/include/EGL/eglext.h
    safe_rm /usr/include/EGL/eglplatform.h
    safe_rm /usr/include/GLES/gl.h
    safe_rm /usr/include/GLES/glext.h
    safe_rm /usr/include/GLES/glplatform.h
    safe_rm /usr/include/GLES2/gl2.h
    safe_rm /usr/include/GLES2/glext2.h
    safe_rm /usr/include/GLES2/glplatform2.h
    safe_rm /usr/include/KHR/khrplatform.h
    safe_rm /usr/include/VG/openvg.h
    safe_rm /usr/include/VG/vgext.h
    safe_rm /usr/include/VG/vgplatform.h
    safe_rm /usr/include/VG/vgu.h

	verbose "$FUNCNAME Default Uninstallation exit"

	return $E_OK
}
#------------------------------------------------------------------------------
# Check for X-Server Version
#------------------------------------------------------------------------------
get_xversion() {
	# Check for the X version
	local xversionFileName=.xversionFile
        X -version &> $INSTALL_DIR/$xversionFileName
	xversion=`cat $xversionFileName | grep "X.Org X Server" | awk '{print $4}'`
	if echo $xversion | grep -q "1.9"; then
		verbose "Found Xorg-xserver-1.9 version"
		XVERSION=$XVERSION_1_9
	else
		verbose "EMGD does not support this X version $xversion"
		verbose "The installer will default to the X version $XVERSION_1_9"
		XVERSION=$XVERSION_1_9
		return $E_OK
	fi
		
}

#------------------------------------------------------------------------------
# Check for OS type
#	Inputs: none
#	Ouputs: none
#------------------------------------------------------------------------------
get_distro() {
	get_xversion
	
	# Check for Fedora 14 
	fedora_version=`cat ${ISSUEFILE} | grep -i "fedora" | awk '{print $3}'`
	timesys_version=`cat ${ISSUEFILE} | grep -i "timesys" | awk '{print $3}'`
	if [[ $fedora_version = "14" ]] || [[ $timesys_version = "14" ]]; then 
		verbose "Found Fedora 14 distribution"
		DISTRO=$DISTRO_FEDORA14
		DISTRO_FOLDER=$DISTRO_F14_FOLDER
		return
	fi

	meego_version=`cat ${ISSUEFILE} | grep -i "MeeGo release" | awk '{print $3}'`
	if [[ $meego_version = "1.1.9" || $meego_version == 1.2* ]]; then
		verbose "Found MeeGo 1.2 distribution"
		DISTRO=$DISTRO_MEEGO1_2
		DISTRO_FOLDER=$DISTRO_MEE1_2_FOLDER
	else
		verbose "Distribution not supported. "
		verbose "The installer will default to the Fedora 14 binaries"
		DISTRO=$DISTRO_UNKNOWN
		DISTRO_FOLDER=$DISTRO_F14_FOLDER
	fi

}
#------------------------------------------------------------------------------
# Check for Chipset
# Checks for the chipset for OpenGL and video decode installation
#	Inputs: none
#	Ouputs: none
#	Returns: E_OK or E_NOTSUPPORTED
#------------------------------------------------------------------------------
get_chipset() {
	# Get chipset information
		chipset=""
		local platform=""

		# There are three libGL binaries included in the package that
		# support different hardware types
		#Check for US15
		if [ ${#chipset} -eq 0 ]; then
			for (( i=0;i<${#EMGD_US15_DEV[@]};i++ )); do
				platform=`$LSPCI_CMD -d "$INTEL_VEN":"${EMGD_US15_DEV[${i}]}"`
				if [ ${#platform} -gt 0 ]; then
					return $E_OK	
				fi
			done
		fi		
		
		#Check for Atom E6xx
		if [ ${#chipset} -eq 0 ]; then
			for (( i=0;i<${#EMGD_TNC_DEV[@]};i++ )); do
				platform=`$LSPCI_CMD -d "$INTEL_VEN":"${EMGD_TNC_DEV[${i}]}"`
				if [ ${#platform} -gt 0 ]; then
					return $E_OK	
				fi
			done
		fi

		return $E_NOTSUPPORTED
}

###############################################################################
#
# Installation Starts Here
#
###############################################################################

#------------------------------------------------------------------------------
# Global Pre-Install
#------------------------------------------------------------------------------

cat <<EOF

Welcome to the $PRODUCT Linux Installer!

This installation script will allow you to install or uninstall $PRODUCT 
on your Linux distribution. 

Supported distributions: 
$DISTRO_FEDORA14

This script will allow you to install or uninstall $PRODUCT, including 
the X driver, OpenGL libraries, PowerVR driver components, and the
EMGD DRM module.

You may need to disable SELinux security to allow the EMGD driver to
load.  Alternatively, you will need to configure SELinux to allow EMGD
to operate.

EOF

# Process command line arguments
while getopts ":uvhrq" OPTION; do
case $OPTION in
	h ) show_usage; exit $E_OK;;
	q ) SILENT=1; verbose "Silent installation";;
	r ) REINSTALL=1;;
	u ) UNINSTALL=1;;
	v ) VERBOSE=1; verbose "Turning on VERBOSE";;
esac
done

verbose "Global Pre-Installation/Uninstallation"

# Which distribution are we running?
get_distro

if [ $? -ne 0 ]; then 
	echo ""
	echo "The X Version is not supported by the EMGD Linux Installer."
	echo ""
	exit $E_NOTSUPPORTED
fi

# Which chipset?
get_chipset

if [ $? -ne 0 ]; then 
	echo ""
	echo "The chipset is not supported by the EMGD Linux Installer."
	echo ""
	exit $E_NOTSUPPORTED
fi

# Check for root level permissions
if [ ! $(id -u) == 0 ]; then
	echo ""
	echo "You need to be root to execute this script."
	echo ""
	exit $E_NOROOT
fi 

# Check to make sure SELinux is disabled
selinuxenabled
if [[ $? -eq 0 && $SILENT -eq 0 ]]; then
	echo ""
	echo -ne "The EMGD Linux Installer has detected that SELinux is enabled.  "
	echo -ne "To continue, SELinux will need to be disabled.  "
	echo -ne "Would you like to continue and disable SELinux? [y/N] "
	read DISABLE_SELINUX
fi

if [[ $DISABLE_SELINUX == "Y" || $DISABLE_SELINUX == "y" ]]; then
	# Disable SELinux
	setenforce 0 
else
	echo ""
	echo "SELinux needs to be disabled to complete EMGD Linux installation."
	echo""
	exit $E_SEENABLED
fi

#------------------------------------------------------------------------------
# Global Install/Uninstallation
#------------------------------------------------------------------------------
verbose "INSTALL_DIR=$INSTALL_DIR"
verbose "DISTRO=$DISTRO"

if [[ $UNINSTALL -eq 1 ]]; then

	echo -ne "Starting uninstall..."	
	# Check if user would like to continue
	# Skip if silent install
	if [[ $SILENT -eq 0 ]]; then
		# If unsupported distro then ask user if they would like to run the standard
		# install
		echo ""
		echo ""		
		if [[ $DISTRO == $DISTRO_UNKNOWN ]]; then
			distro_input="N"
			echo -ne "Unsupported Linux distribution."
		fi
		echo -ne "Would you like to continue with the uninstallation? [y/N] "
			read distro_input

			if [[ $distro_input != "Y" && $distro_input != "y" ]]; then
				exit $E_OK
			fi		
	fi

    if [[ $DISTRO == $DISTRO_MEEGO1_2 ]]; then
        echo "EMGD is a builtin component of the Meego 1.2 distribution, so"
        echo "uninstallation is not supported by this script.  Userspace"
        echo "driver libraries may be removed by uninstalling the 'emgd-bin'"
        echo "RPM via the packaging system, although doing so will render"
        echo "the system unusable unless another graphics driver with OpenGL-ES"
        echo "support is installed."
        exit $E_NOTSUPPORTED
    fi

    default_uninstall

	# Uninstall the EMGD DRM module 
	echo "Uninstalling the EMGD DRM module"
	cd "$INSTALL_DIR/../$DISTRO_FOLDER/driver"
	
	verbose "Global Post-Uninstallation"

	# Do DRM Post-Installation
	$MODPROBE_CMD -r emgd 

	# Remove the module completely
	KERNEL=`uname -r`
	safe_rm /lib/modules/$KERNEL/kernel/drivers/gpu/drm/emgd/emgd.ko

	# Done
	cd $INSTALL_DIR

	echo ""
	echo "Completed $PRODUCT Uninstallation"
	echo ""
else
		
	# Agree to terms of license
	verbose "INSTALL_DIR=$INSTALL_DIR"
	cd "$INSTALL_DIR/../License"
	if [ -e "License.txt" ]; then
		cat License.txt
	else
		cat pre_license.txt
	fi
	
	# Skip license agreement and unsupported distro check if silent install
	if [[ $SILENT -eq 0 ]]; then
		# Check if user would like to continue
		install_input="N"
		echo ""
		echo ""
		echo -ne "Do you agree to the terms of the license agreement? [y/N] "
		read install_input

		if [[ $install_input != "Y" && $install_input != "y" ]]; then
			exit $E_OK
		fi

		# If unsupported distro then ask user if they would like to run the standard
		# install
		if [[ $DISTRO == $DISTRO_UNKNOWN ]]; then
			distro_input="N"
			echo ""
			echo ""
			echo -ne "Unsupported Linux distribution."
			echo -ne "Would you like to continue with the installation anyway? [y/N] "
			read distro_input

			if [[ $distro_input != "Y" && $distro_input != "y" ]]; then
				exit $E_OK
			fi
		fi
	fi	

	# This is a standard install
	pre_install
	
	# Verify that the pre-install succesfully completed
	if [ $? -ne 0 ]; then 
		echo ""
		echo "The EMGD Linux Installer is exiting."
		echo ""
		exit $E_NODIR
	fi

    if [[ $DISTRO == $DISTRO_MEEGO1_2 ]]; then
        echo "EMGD is a builtin component of the Meego 1.2 distribution, so"
        echo "direct upgrading/installation via this script is not supported."
        echo "Please use the RPM installation/upgrade method described in the"
        echo "EMGD user's guide."
        exit $E_NOTSUPPORTED
    fi

	copy_files
	create_links
	emgd_gui

	# Check to make sure INSMOD command exists
	if [ ! -e $INSMOD_CMD ]; then
		echo "The command $INSMOD_CMD not found"
		INSMOD_CMD=""	
	fi
	
	# Check to make sure LSMOD command exists
	if [ ! -e $LSMOD_CMD ]; then
		echo "The command $LSMOD_CMD not found, exiting."
		exit $E_NOCMD
	fi

	drm

	if [ $? -ne 0 ]; then 
		echo ""
		echo "The EMGD DRM build reported errors. "
		echo "The EMGD Linux Installer is exiting."
		echo ""
		exit $E_NOTSUPPORTED
	fi

	cd $INSTALL_DIR

	echo ""
	echo "Completed $PRODUCT Installation"
	echo ""
fi

exit $E_OK
